from typing import List, Any

from graphql import GraphQLError, GraphQLSchema, DocumentNode, ValidationContext, visit, TypeInfo

from graphql_utilities.exceptions import ValidationAbortedError
from graphql_utilities.visitor import DepthAnalysisVisitor, CostAnalysisVisitor


def validate_depth(
        schema: GraphQLSchema,
        document: DocumentNode,
        context_value: Any,
        errors: List[GraphQLError]
):
    def on_error(error: GraphQLError) -> None:
        errors.append(error)
        raise ValidationAbortedError

    depth_analysis = context_value.get("depth_analysis")

    if depth_analysis:
        max_depth = depth_analysis.get("max_depth", 10)
        context = ValidationContext(schema=schema, ast=document, type_info=TypeInfo(schema), on_error=on_error)

        visit(document, DepthAnalysisVisitor(context=context, max_depth=max_depth))


def validate_cost(
        schema: GraphQLSchema,
        document: DocumentNode,
        context_value: Any,
        errors: List[GraphQLError]
):
    def on_error(error: GraphQLError) -> None:
        errors.append(error)
        raise ValidationAbortedError

    cost_analysis = context_value.get("cost_analysis")

    if cost_analysis:
        max_complexity = cost_analysis.get("max_complexity", 500)
        context = ValidationContext(schema=schema, ast=document, type_info=TypeInfo(schema), on_error=on_error)

        visit(document, CostAnalysisVisitor(context=context, max_complexity=max_complexity))

    return []



from django.conf import settings
from django.core.handlers.wsgi import WSGIRequest

from graphql import execute, parse

from graphql_utilities.decorators import run_only_once
from tests.helpers import assert_no_errors
from tests.schema import schema

settings.configure(name="TEST")


def describe_run_only_once():
    query = '{ field_1_str field_2_int field_3_obj { field_3_obj_sub_1 { xxx } } }'

    def test_run_only_once():
        class OneShotMiddleware:
            @run_only_once
            def resolve(self, next_, *args, **kwargs):
                runs.append("OneShotMiddleware")
                return next_(*args, **kwargs)

        runs = []
        middleware = (OneShotMiddleware(),)

        result = execute(schema=schema, document=parse(query),
                         middleware=middleware, context_value={})
        assert_no_errors(result)
        assert len(runs) == 1, "middleware should run only once"

    def test_run_only_once_doesnt_affect_other_middleware():
        runs = []

        class OneShotMiddleware:
            @run_only_once
            def resolve(self, next_, *args, **kwargs):
                runs.append("OneShotMiddleware")
                return next_(*args, **kwargs)

        class FieldMiddleware:
            def resolve(self, next_, *args, **kwargs):
                runs.append("FieldMiddleware")
                return next_(*args, **kwargs)

        result = execute(schema=schema, document=parse(query),
                         middleware=(FieldMiddleware(), OneShotMiddleware()),
                         context_value={})
        assert_no_errors(result)
        assert len(runs) == 5
        assert runs.count("OneShotMiddleware") == 1
        assert runs.count("FieldMiddleware") == 4

    def test_middleware_run_on_multiple_requests():
        runs = []

        class OneShotMiddleware:
            @run_only_once
            def resolve(self, next_, *args, **kwargs):
                runs.append("OneShotMiddleware")
                return next_(*args, **kwargs)

        middleware = (OneShotMiddleware(),)
        [execute(schema=schema, document=parse(query), middleware=middleware, context_value={}) for _ in (0, 1)]
        assert len(runs) == 2, "middleware should run twice for 2 requests"

    def test_middleware_with_graphene_context_values():
        runs = []

        class Context(object):
            # This class is taken from Graphene source code
            def __init__(self, **params):
                for key, value in params.items():
                    setattr(self, key, value)

        class OneShotMiddleware:
            @run_only_once
            def resolve(self, next_, *args, **kwargs):
                runs.append("OneShotMiddleware")
                return next_(*args, **kwargs)

        middleware = (OneShotMiddleware(),)
        [execute(schema=schema, document=parse(query), middleware=middleware, context_value=Context()) for _ in (0, 1)]
        assert len(runs) == 2, "middleware should run twice for 2 requests"

    def test_middleware_with_wsgi_request_context_values():
        runs = []

        class OneShotMiddleware:
            @run_only_once
            def resolve(self, next_, *args, **kwargs):
                runs.append("OneShotMiddleware")
                return next_(*args, **kwargs)

        middleware = (OneShotMiddleware(),)
        [execute(schema=schema, document=parse(query), middleware=middleware,
                 context_value=_get_fake_wsgi_request_obj()) for _ in (0, 1)]
        assert len(runs) == 2, "middleware should run twice for 2 requests"

    def test_multiple_one_shot_middleware():
        runs = []

        class OneShotMiddleware:
            @run_only_once
            def resolve(self, next_, *args, **kwargs):
                runs.append("OneShotMiddleware")
                return next_(*args, **kwargs)

        class AnotherOneShotMiddleware:
            @run_only_once
            def resolve(self, next_, *args, **kwargs):
                runs.append("AnotherOneShotMiddleware")
                return next_(*args, **kwargs)

        middleware = (OneShotMiddleware(), AnotherOneShotMiddleware())
        execute(schema=schema, document=parse(query), middleware=middleware,
                context_value={})
        assert len(runs) == 2, "both middleware should have run"
        assert "OneShotMiddleware" in runs
        assert "AnotherOneShotMiddleware" in runs

    def _get_fake_wsgi_request_obj():
        return WSGIRequest({
            'REQUEST_METHOD': 'POST',
            'PATH_INFO': '/graphql',
            'wsgi.input': ""})



def run_only_once(resolve_func):
    """
    Make sure middleware is run only once,
    this is done by setting a flag in the `context` of `ResolverInfo`

    Example:
        class AuthenticationMiddleware:
            @run_only_once
            def resolve(self, next, root, info, *args, **kwargs):
                pass
    """
    from functools import wraps

    @wraps(resolve_func)
    def wrapper(self, next, root, info, *args, **kwargs):
        has_context = info.context is not None
        decorator_name = "__{}_run__".format(self.__class__.__name__)

        if has_context:
            if isinstance(info.context, dict) and not info.context.get(decorator_name, False):
                info.context[decorator_name] = True
                return resolve_func(self, next, root, info, *args, **kwargs)
            elif not isinstance(info.context, dict) and not getattr(info.context, decorator_name, False):
                # Graphene: it could be a Context or WSGIRequest object
                setattr(info.context, decorator_name, True)
                return resolve_func(self, next, root, info, *args, **kwargs)

        # No context, run_only_once will not work
        return next(root, info, *args, **kwargs)

    return wrapper



from typing import Union

from graphql import Source, build_schema as _build_schema

from graphql_utilities.directives import cost_directive_source_doc


def build_schema_with_cost(
        source: Union[str, Source],
        assume_valid=False,
        assume_valid_sdl=False,
        no_location=False,
        experimental_fragment_variables=False):
    return _build_schema(source=cost_directive_source_doc + source, assume_valid=assume_valid,
                         assume_valid_sdl=assume_valid_sdl, no_location=no_location,
                         experimental_fragment_variables=experimental_fragment_variables)



from graphql import GraphQLSchema, GraphQLObjectType, GraphQLField, GraphQLString, GraphQLInt

from graphql_utilities import build_schema_with_cost

schema = GraphQLSchema(
    query=GraphQLObjectType(
        name='RootQueryType',
        fields={
            'field_1_str': GraphQLField(
                GraphQLString,
                resolve=lambda obj, info: 'str_1_field'),
            'field_2_int': GraphQLField(
                GraphQLInt,
                resolve=lambda obj, info: 2),
            'field_3_obj': GraphQLField(
                GraphQLObjectType(
                    name='Field3Object',
                    fields={
                        'field_3_obj_sub_1': GraphQLField(GraphQLObjectType(
                            name='Field3ObjectSubfield1',
                            fields={
                                'xxx': GraphQLField(GraphQLString)
                            }
                        ))
                    }
                ),
                resolve=lambda obj, info: {'field_3_obj': {'field_3_obj_sub_1': '1_sub_obj_3_field'}}
            ),
            'cat': GraphQLField(
                GraphQLObjectType(
                    name='CatObject',
                    fields={
                        'meow': GraphQLField(GraphQLString)
                    },
                ),
                resolve=lambda obj, info: {"meow": "woem"}
            )
        }))

post_schema = build_schema_with_cost("""
type Author @cost(complexity: 4) {
    uid: String!
    name: String @cost(complexity: 2)
    email: String @cost(complexity: 2)
}

interface TimestampedType {
    createdAt: String @cost(complexity: 2)
    updatedAt: String @cost(complexity: 2)
}

type Post implements TimestampedType @cost(complexity: 8) {
    author: Author
    postId: String! @cost(complexity: 4)
    title: String @cost(complexity: 2)
    text: String @cost(complexity: 1)
    publishedAt: String @cost(complexity: 1)
    isPublic: Boolean  # NO cost
    createdAt: String @cost(complexity: 4) # OVERRIDE cost
    updatedAt: String
}

type Announcement implements TimestampedType {
    createdAt: String
    updatedAt: String
    announcementId: String! @cost(complexity: 4)
    title: String
    text: String
}

union PostOrAnnouncement = Post | Announcement

type Query {
    posts(first: Int): [Post]  @cost(multipliers: ["first"])
    postsWithOverride(first: Int): [Post]  @cost(complexity: 2, multipliers: ["first"]) # OVERRIDE complexity of Post type
    postsOrAnnouncements(first: Int): [PostOrAnnouncement] @cost(multipliers: ["first"])
    post(id: ID!): Post
    announcement(id: ID!): Announcement
}
""")



from functools import reduce
from typing import Tuple, List, Union

from graphql import Visitor, GraphQLError, BREAK, IDLE, ValidationContext, GraphQLInterfaceType, GraphQLObjectType, \
    GraphQLField, get_named_type
from graphql.execution.values import get_argument_values, get_directive_values
from graphql.language import ast

from graphql_utilities.helpers import deep_get


class DepthAnalysisVisitor(Visitor):
    def __init__(self, context: ValidationContext, max_depth: int = 10):
        self.max_depth = max_depth
        self.context = context

    def enter_name(self, node, key, parent, path, *args):
        if isinstance(node, ast.NameNode):
            depth = path.count("selection_set")
            if depth > self.max_depth:
                self.context.report_error(GraphQLError(f"Reached max depth of {self.max_depth}"))
                return BREAK

        return IDLE


class CostAnalysisVisitor(Visitor):
    def __init__(self, context: ValidationContext, max_complexity):
        self.context = context
        self.max_complexity = max_complexity
        self.operation_multipliers = []
        self.total_complexity = 0  # Keep track of complexity so far

    def enter_operation_definition(self, node, key, parent, path, *args):
        """
        Inspired by:
        https://github.com/pa-bru/graphql-cost-analysis/blob/b0f41119955654b52f8943ceaf4eb35dfb68db4e/src/costAnalysis.js#L95
        """
        cost_compute_fn = self.get_cost_compute_fn(operation_type=node.operation.name)
        if not cost_compute_fn:
            return BREAK

        self.total_complexity += cost_compute_fn(node=node)

        return IDLE

    def exit_operation_definition(self):
        """
        Ref:
        https://github.com/pa-bru/graphql-cost-analysis/blob/b0f41119955654b52f8943ceaf4eb35dfb68db4e/src/costAnalysis.js#L124
        """
        if self.total_complexity > self.max_complexity:
            return self.context.report_error(
                GraphQLError(f"{self.total_complexity} exceeded max complexity of {self.max_complexity}"))

    def get_cost_compute_fn(self, operation_type):
        fn = getattr(self, f"compute_{operation_type.lower()}_cost")

        if not fn:
            self.context.report_error(
                GraphQLError(f"Query cost could not be calculated for operation of type {operation_type}"))

        return fn

    def compute_query_cost(self, node):
        return self.compute_node_cost(
            node=node,
            type_definition=self.context.schema.query_type
        )

    def compute_mutation_cost(self, node):
        return self.compute_node_cost(
            node=node,
            type_definition=self.context.schema.mutation_type
        )

    def compute_subscription_cost(self, node):
        return self.compute_node_cost(
            node=node,
            type_definition=self.context.schema.subscription_type
        )

    def compute_node_cost(self, node, type_definition, parent_multiplier=None, parent_complexity=None):
        if not parent_multiplier:
            parent_multiplier = []

        if not node.selection_set:
            return 0

        fields = {}
        if isinstance(type_definition, GraphQLObjectType) or isinstance(type_definition, GraphQLInterfaceType):
            fields = type_definition.fields

        total_cost = 0
        fragment_costs = []
        variables = {}  # TODO get variables from operation
        selections = node.selection_set.selections
        for selection in selections:
            self.operation_multipliers = [*parent_multiplier]
            node_cost = 0

            if selection.kind == 'field':
                # Calculate cost for FieldNode
                field: GraphQLField = fields.get(selection.name.value)
                if not field:
                    break

                field_type = get_named_type(field.type)
                field_args = get_argument_values(field, selection, variables)

                use_field_type_complexity = False
                cost_is_computed = False
                if field.ast_node and field.ast_node.directives:
                    directive_args: Union[Tuple[int, List, bool], None] = self.get_args_from_directives(
                        directives=field.ast_node.directives,
                        field_args=field_args
                    )

                    override_complexity = directive_args[-1]
                    if not override_complexity and isinstance(field_type, GraphQLObjectType):
                        use_field_type_complexity = True
                        parent_complexity, _, _ = self.get_args_from_directives(
                            directives=field_type.ast_node.directives,
                            field_args=field_args)

                    node_cost = self.compute_cost(directive_args)

                    if directive_args:
                        cost_is_computed = True

                if field_type and field_type.ast_node and \
                        field_type.ast_node.directives and \
                        isinstance(field_type, GraphQLObjectType) and \
                        (not cost_is_computed or use_field_type_complexity):
                    directive_args = self.get_args_from_directives(directives=field_type.ast_node.directives,
                                                                   field_args=field_args)
                    node_cost = self.compute_cost(directive_args)

                child_cost = self.compute_node_cost(node=selection, type_definition=field_type,
                                                    parent_multiplier=self.operation_multipliers,
                                                    parent_complexity=parent_complexity) or 0
                node_cost += child_cost

            elif selection.kind == 'fragment_spread':
                fragment = self.context.get_fragment(selection.name.value)
                fragment_type = fragment and self.context.schema.get_type(fragment.type_condition.name.value)
                fragment_node_cost = self.compute_node_cost(fragment, fragment_type, self.operation_multipliers) \
                    if fragment \
                    else 0
                fragment_costs.append(fragment_node_cost)
                node_cost = 0

            elif selection.kind == 'inline_fragment':
                inline_fragment_type = self.context.schema.get_type(selection.type_condition.name.value) \
                    if selection.type_condition and selection.type_condition.name \
                    else type_definition
                fragment_node_cost = self.compute_node_cost(selection, inline_fragment_type, self.operation_multipliers) \
                    if selection \
                    else 0

                fragment_costs.append(fragment_node_cost)
                node_cost = 0

            else:
                node_cost = self.compute_node_cost(node=selection, type_definition=type_definition)

            total_cost += max(node_cost, 0)

        if fragment_costs:
            return total_cost + max(fragment_costs)

        return total_cost

    def get_args_from_directives(self, directives, field_args) -> Union[None, Tuple[Union[int, None], List, bool]]:
        cost_directive = None
        cost_directives = list(filter(lambda d: d.name.value == 'cost', directives))
        if len(cost_directives) > 0:
            cost_directive = cost_directives[0]

        if cost_directive and cost_directive.arguments:
            complexity_arg = None
            complexity_args = list(filter(lambda a: a.name.value == 'complexity', cost_directive.arguments))
            if len(complexity_args) > 0:
                complexity_arg = complexity_args[0]

            multipliers_arg = None
            multipliers_args = list(filter(lambda a: a.name.value == 'multipliers', cost_directive.arguments))
            if len(multipliers_args) > 0:
                multipliers_arg = multipliers_args[0]

            has_complexity_defined = bool(
                complexity_arg and complexity_arg.value and complexity_arg.value.kind == "int_value")
            complexity = int(complexity_arg.value.value) if has_complexity_defined else 0

            has_multipliers_defined = multipliers_arg and multipliers_arg.value and multipliers_arg.value.kind == "list_value"
            multipliers = self.get_multipliers_from_list_node(multipliers_arg.value.values, field_args) \
                if has_multipliers_defined else []

            override_complexity = has_complexity_defined  # complexity is set in @cost(), override @cost() of Object

            return complexity, multipliers, override_complexity

    def compute_cost(self, directive_args: Tuple[int, List, bool]) -> int:
        [complexity, multipliers, override_complexity] = directive_args

        use_multipliers = len(multipliers) > 0 or len(self.operation_multipliers) > 0
        if use_multipliers:
            if len(multipliers) > 0:
                multiplier = reduce(
                    lambda total, current: total + current,
                    multipliers,
                    0
                )
                self.operation_multipliers = [*self.operation_multipliers, multiplier]

            return reduce(
                lambda acc, mltp: acc * mltp,
                self.operation_multipliers,
                complexity
            )
        return complexity

    def get_multipliers_from_list_node(self, list_nodes, field_args) -> List[int]:
        multipliers = list(map(lambda node: node.value, filter(lambda node: node.kind == "string_value", list_nodes)))
        return self.get_multipliers_from_string(multipliers, field_args)

    def get_multipliers_from_string(self, multipliers, field_args) -> List[int]:
        def multiplier_selector(multiplier):
            value = deep_get(dictionary=field_args, keys=multiplier)
            if isinstance(value, list):
                return len(value)
            return int(value or 0)

        mapped_values = map(multiplier_selector, multipliers)
        filtered_values = filter(lambda multiplier: multiplier != 0, mapped_values)
        return list(filtered_values)



from unittest.mock import Mock

from graphql import TypeInfo, visit, ValidationContext, parse

from graphql_utilities.visitor import CostAnalysisVisitor
from tests.schema import post_schema


def describe_cost_analysis():
    on_error_stub = Mock()

    def _calculate_cost(ast_document, max_complexity=100) -> int:
        context = ValidationContext(schema=post_schema, ast=ast_document,
                                    type_info=TypeInfo(post_schema), on_error=on_error_stub)
        visitor = CostAnalysisVisitor(context=context, max_complexity=max_complexity)
        visit(ast_document, visitor)
        return visitor.total_complexity

    def test_simple_scalar_only_cost():
        document = parse("""
            query {
                announcement(id: "AAAAA") {
                    announcementId
                }
            }
        """)
        complexity = _calculate_cost(ast_document=document)
        assert complexity == 4

    def test_simple_scalar_cost():
        # Directives are not supposed to be inherited
        document = parse("""
                    query {
                        announcement(id: "AAAAA") {
                            createdAt
                            announcementId
                        }
                    }
                """)
        complexity = _calculate_cost(ast_document=document)
        assert complexity == 4

    def test_simple_cost_with_object_type():
        document = parse("""
            query {
                post(id: "XXXXXXXXXXXXXX") {
                    postId
                    title
                    author {
                        name
                    }
                }
            }
        """)
        complexity = _calculate_cost(ast_document=document)
        assert complexity == 20

    def test_overridden_cost():
        # Directives are not supposed to be inherited
        document = parse("""
            query {
                post(id: "XXXXXXXXXXXXXX") {
                    postId
                    createdAt
                    updatedAt
                }
            }
        """)
        complexity = _calculate_cost(ast_document=document)
        assert complexity == 16

    def test_multiplier_without_overriding_complexity():
        document = parse("""
            query {
                posts(first: 5) {
                    title
                    isPublic
                }
            }
        """)

        complexity = _calculate_cost(ast_document=document)
        assert complexity == 50

    def test_multiplier_with_complexity_overridden():
        document = parse("""
                    query {
                        postsWithOverride(first: 5) {
                            title
                            isPublic
                        }
                    }
                """)

        complexity = _calculate_cost(ast_document=document)
        assert complexity == 20

    def test_inline_fragment():
        document = parse("""
            query {
                postsOrAnnouncements(first: 5) {
                    ... on Post {
                        postId
                    }
                    ... on Announcement {
                        announcementId
                    }
                }
            }
        """)

        complexity = _calculate_cost(ast_document=document)
        assert complexity == 20

    def test_fragment_spread():
        document = parse("""
                    query {
                        postsOrAnnouncements(first: 5) {
                            ... postIdField
                            ... announcementIdField
                        }
                    }

                    fragment announcementIdField on Announcement {
                        announcementId
                    }

                    fragment postIdField on Post {
                        postId
                        createdAt
                    }
                """)

        complexity = _calculate_cost(ast_document=document)
        assert complexity == 40



from .decorators import run_only_once
from .directives import GraphQLCostDirective, schema_with_cost_directive, cost_directive_source_doc
from .execution import ExtendedExecutionContext
from .utilities import build_schema_with_cost

__version__ = "0.4.0"

__all__ = [
    "run_only_once",
    "ExtendedExecutionContext",
    "GraphQLCostDirective",
    "schema_with_cost_directive",
    "cost_directive_source_doc",
    "build_schema_with_cost"
]



from typing import Any, Dict, Union, List, Optional

from graphql import GraphQLError, ExecutionContext, GraphQLSchema, DocumentNode, \
    GraphQLFieldResolver, GraphQLTypeResolver, Middleware

from graphql_utilities.exceptions import ValidationAbortedError
from graphql_utilities.validate import validate_depth, validate_cost

ContextValue = Optional[Dict[str, Any]]


class ExtendedExecutionContext(ExecutionContext):
    @classmethod
    def build(
            cls,
            schema: GraphQLSchema,
            document: DocumentNode,
            root_value: Any = None,
            context_value: Any = None,
            raw_variable_values: Dict[str, Any] = None,
            operation_name: str = None,
            field_resolver: GraphQLFieldResolver = None,
            type_resolver: GraphQLTypeResolver = None,
            middleware: Middleware = None,
    ) -> Union[List[GraphQLError], ExecutionContext]:
        # As suggested, graphql_sync, graphql_impl, or graphql should be invoked to execute queries.
        # By invoking them, schema and documents will have already been validated before the execution
        # context is being created
        errors: List[GraphQLError] = []

        if context_value and isinstance(context_value, dict):
            try:
                validate_depth(schema=schema, document=document, context_value=context_value, errors=errors)
                validate_cost(schema=schema, document=document, context_value=context_value, errors=errors)
            except ValidationAbortedError:
                return errors

        return super(ExtendedExecutionContext, cls).build(
            schema,
            document,
            root_value,
            context_value,
            raw_variable_values,
            operation_name,
            field_resolver,
            type_resolver,
            middleware
        )



"""
Directives

Sample source code:
    GraphQLIncludeDirective = GraphQLDirective(
        name="include",
        locations=[
            DirectiveLocation.FIELD,
            DirectiveLocation.FRAGMENT_SPREAD,
            DirectiveLocation.INLINE_FRAGMENT,
        ],
        args={
            "if": GraphQLArgument(
                GraphQLNonNull(GraphQLBoolean), description="Included when true."
            )
        },
        description="Directs the executor to include this field or fragment"
        " only when the `if` argument is true.",
    )

    How to use?
        skip = get_directive_values(GraphQLSkipDirective, node, self.variable_values)

Ref:
- https://github.com/graphql-python/graphql-core-next/issues/65
- http://spec.graphql.org/June2018/#sec-Type-System.Directives
"""
from graphql import build_schema

cost_directive_source_doc = """
    directive @cost(
        complexity: Int
        multipliers: [String]
    ) on FIELD_DEFINITION | OBJECT
"""

schema_with_cost_directive = build_schema(source=cost_directive_source_doc, no_location=True)

GraphQLCostDirective = schema_with_cost_directive.get_directive("cost")
